'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});

var _symbol = require('babel-runtime/core-js/symbol');

var _symbol2 = _interopRequireDefault(_symbol);

var _assign = require('babel-runtime/core-js/object/assign');

var _assign2 = _interopRequireDefault(_assign);

var _map = require('babel-runtime/core-js/map');

var _map2 = _interopRequireDefault(_map);

var _classCallCheck2 = require('babel-runtime/helpers/classCallCheck');

var _classCallCheck3 = _interopRequireDefault(_classCallCheck2);

var _createClass2 = require('babel-runtime/helpers/createClass');

var _createClass3 = _interopRequireDefault(_createClass2);

exports.default = defaultCellRangeRenderer;

var _types = require('./types');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function foreachKey(rowStartIndex, rowStopIndex, columnStartIndex, columnStopIndex, callback) {
  for (var rowIndex = rowStartIndex; rowIndex <= rowStopIndex; rowIndex++) {
    for (var columnIndex = columnStartIndex; columnIndex <= columnStopIndex; columnIndex++) {
      var key = getKey(rowIndex, columnIndex);
      callback(key);
    }
  }
}

var ReusableKeyCache = function () {
  function ReusableKeyCache() {
    (0, _classCallCheck3.default)(this, ReusableKeyCache);

    this.lastMap = new _map2.default();
  }

  (0, _createClass3.default)(ReusableKeyCache, [{
    key: 'getReusableKeyMap',
    value: function getReusableKeyMap(rowStartIndex, rowStopIndex, columnStartIndex, columnStopIndex) {
      var newMap = new _map2.default();
      var usedKeys = [];
      var lastMap = this.lastMap;
      var nextKey = 1;

      foreachKey(rowStartIndex, rowStopIndex, columnStartIndex, columnStopIndex, function (key) {
        var lastKey = lastMap.get(key);
        if (lastKey != null) {
          newMap.set(key, lastKey);
          usedKeys[lastKey] = true;
        }
      });

      foreachKey(rowStartIndex, rowStopIndex, columnStartIndex, columnStopIndex, function (key) {
        var newKey = newMap.get(key);
        if (newKey == null) {
          // find new key for this
          while (newKey == null) {
            if (!usedKeys[nextKey]) {
              newKey = nextKey;
            }
            nextKey++;
          }
        }
        newMap.set(key, newKey);
        usedKeys[newKey] = true;
      });

      this.lastMap = newMap;
      return newMap;
    }
  }], [{
    key: 'getReusableKeyCache',
    value: function getReusableKeyCache(cacheOnThisObject) {
      var reusableKeyCache = cacheOnThisObject[ReusableKeyCache.symbol];
      if (reusableKeyCache == null) {
        reusableKeyCache = cacheOnThisObject[ReusableKeyCache.symbol] = new ReusableKeyCache();
      }
      return reusableKeyCache;
    }
  }]);
  return ReusableKeyCache;
}();

ReusableKeyCache.symbol = (0, _symbol2.default)('ReusableKeyCache_symbol');


function getKey(rowIndex, columnIndex) {
  return rowIndex + '-' + columnIndex;
}

/**
 * Default implementation of cellRangeRenderer used by Grid.
 * This renderer supports cell-caching while the user is scrolling.
 */

function defaultCellRangeRenderer(_ref) {
  var cellCache = _ref.cellCache,
      cellRenderer = _ref.cellRenderer,
      columnSizeAndPositionManager = _ref.columnSizeAndPositionManager,
      columnStartIndex = _ref.columnStartIndex,
      columnStopIndex = _ref.columnStopIndex,
      deferredMeasurementCache = _ref.deferredMeasurementCache,
      horizontalOffsetAdjustment = _ref.horizontalOffsetAdjustment,
      isScrolling = _ref.isScrolling,
      isScrollingOptOut = _ref.isScrollingOptOut,
      parent = _ref.parent,
      rowSizeAndPositionManager = _ref.rowSizeAndPositionManager,
      rowStartIndex = _ref.rowStartIndex,
      rowStopIndex = _ref.rowStopIndex,
      styleCache = _ref.styleCache,
      verticalOffsetAdjustment = _ref.verticalOffsetAdjustment,
      visibleColumnIndices = _ref.visibleColumnIndices,
      visibleRowIndices = _ref.visibleRowIndices,
      overscanCellCount = _ref.overscanCellCount;

  var renderedCells = [];

  var reusableKeyCache = ReusableKeyCache.getReusableKeyCache(parent);
  var reusableKeyMap = reusableKeyCache.getReusableKeyMap(rowStartIndex, rowStopIndex, columnStartIndex, columnStopIndex);

  // Browsers have native size limits for elements (eg Chrome 33M pixels, IE 1.5M pixes).
  // User cannot scroll beyond these size limitations.
  // In order to work around this, ScalingCellSizeAndPositionManager compresses offsets.
  // We should never cache styles for compressed offsets though as this can lead to bugs.
  // See issue #576 for more.
  var areOffsetsAdjusted = columnSizeAndPositionManager.areOffsetsAdjusted() || rowSizeAndPositionManager.areOffsetsAdjusted();

  var canCacheStyle = !isScrolling && !areOffsetsAdjusted;
  var dataOrder = overscanCellCount;

  for (var rowIndex = rowStartIndex; rowIndex <= rowStopIndex; rowIndex++) {
    var rowDatum = rowSizeAndPositionManager.getSizeAndPositionOfCell(rowIndex);

    for (var columnIndex = columnStartIndex; columnIndex <= columnStopIndex; columnIndex++) {
      var columnDatum = columnSizeAndPositionManager.getSizeAndPositionOfCell(columnIndex);
      var isVisible = columnIndex >= visibleColumnIndices.start && columnIndex <= visibleColumnIndices.stop && rowIndex >= visibleRowIndices.start && rowIndex <= visibleRowIndices.stop;
      var key = getKey(rowIndex, columnIndex);
      var reuseKey = reusableKeyMap.get(key).toString();
      var style = void 0;

      // Cache style objects so shallow-compare doesn't re-render unnecessarily.
      if (canCacheStyle && styleCache[key]) {
        style = styleCache[key];
      } else {
        // In deferred mode, cells will be initially rendered before we know their size.
        // Don't interfere with CellMeasurer's measurements by setting an invalid size.
        if (deferredMeasurementCache && !deferredMeasurementCache.has(rowIndex, columnIndex)) {
          // Position not-yet-measured cells at top/left 0,0,
          // And give them width/height of 'auto' so they can grow larger than the parent Grid if necessary.
          // Positioning them further to the right/bottom influences their measured size.
          style = {
            height: 'auto',
            left: 0,
            position: 'absolute',
            top: 0,
            width: 'auto'
          };
        } else {
          style = {
            height: rowDatum.size,
            left: columnDatum.offset + horizontalOffsetAdjustment,
            position: 'absolute',
            top: rowDatum.offset + verticalOffsetAdjustment,
            width: columnDatum.size
          };

          styleCache[key] = style;
        }
      }

      var cellRendererParams = {
        columnIndex: columnIndex,
        isScrolling: isScrolling,
        isVisible: isVisible,
        key: reuseKey,
        parent: parent,
        rowIndex: rowIndex,
        style: style,
        dataOrder: dataOrder++
      };

      var renderedCell = void 0;

      // Avoid re-creating cells while scrolling.
      // This can lead to the same cell being created many times and can cause performance issues for "heavy" cells.
      // If a scroll is in progress- cache and reuse cells.
      // This cache will be thrown away once scrolling completes.
      // However if we are scaling scroll positions and sizes, we should also avoid caching.
      // This is because the offset changes slightly as scroll position changes and caching leads to stale values.
      // For more info refer to issue #395
      //
      // If isScrollingOptOut is specified, we always cache cells.
      // For more info refer to issue #1028
      if ((isScrollingOptOut || isScrolling) && !horizontalOffsetAdjustment && !verticalOffsetAdjustment) {
        if (!cellCache[key]) {
          cellCache[key] = cellRenderer(cellRendererParams);
        }

        renderedCell = cellCache[key];
        if (renderedCell && renderedCell.key !== reuseKey) {
          // this prevents a previously cached cell from using a wrong and possibly duplicate key
          renderedCell = (0, _assign2.default)({}, renderedCell, { key: reuseKey });
        }

        // If the user is no longer scrolling, don't cache cells.
        // This makes dynamic cell content difficult for users and would also lead to a heavier memory footprint.
      } else {
        renderedCell = cellRenderer(cellRendererParams);
      }

      if (renderedCell == null || renderedCell === false) {
        continue;
      }

      if (process.env.NODE_ENV !== 'production') {
        warnAboutMissingStyle(parent, renderedCell);
      }

      renderedCells.push(renderedCell);
    }
  }

  renderedCells.sort(function (a, b) {
    return parseInt(a.key) - parseInt(b.key);
  });

  return renderedCells;
}

function warnAboutMissingStyle(parent, renderedCell) {
  if (process.env.NODE_ENV !== 'production') {
    if (renderedCell) {
      // If the direct child is a CellMeasurer, then we should check its child
      // See issue #611
      if (renderedCell.type && renderedCell.type.__internalCellMeasurerFlag) {
        renderedCell = renderedCell.props.children;
      }

      if (renderedCell && renderedCell.props && renderedCell.props.style === undefined && parent.__warnedAboutMissingStyle !== true) {
        parent.__warnedAboutMissingStyle = true;

        console.warn('Rendered cell should include style property for positioning.');
      }
    }
  }
}